/*
 * Copyright (C) Cross The Road Electronics.  All rights reserved.
 * License information can be found in CTRE_LICENSE.txt
 * For support and suggestions contact support@ctr-electronics.com or file
 * an issue tracker at https://github.com/CrossTheRoadElec/Phoenix-Releases
 */
#pragma once

#include "ctre/phoenix6/configs/Configs.hpp"
#include "ctre/phoenix6/controls/ControlRequest.hpp"
#include "ctre/phoenix6/networking/interfaces/Control_Interface.h"
#include <sstream>

#include <units/frequency.h>
#include <units/time.h>


namespace ctre {
namespace phoenix6 {
namespace controls {

/**
 * Request coast neutral output of actuator.  The bridge is disabled and the
 * rotor is allowed to coast.
 */
class CoastOut : public ControlRequest
{
    bool ApplyConfigsOnRequest{false};

    ctre::phoenix::StatusCode SendRequest(const char *network, uint32_t deviceHash, bool cancelOtherRequests, std::shared_ptr<ControlRequest> &req)
    {
        std::stringstream ss;


        if (req.get() != this)
        {
            auto const reqCast = dynamic_cast<CoastOut *>(req.get());
            if (reqCast != nullptr)
            {
                *reqCast = *this;
            }
            else
            {
                req = std::make_shared<CoastOut>(*this);
            }
        }

        
        std::string strs{ss.str()};
        c_ctre_phoenix6_requestConfigApply(network, deviceHash, ConfigTimeout.to<double>(), strs.c_str(), strs.length(), ApplyConfigsOnRequest);
        ApplyConfigsOnRequest = false;
        return c_ctre_phoenix6_RequestControlCoastOut(network, deviceHash, UpdateFreqHz.to<double>(), cancelOtherRequests);
    }

public:
    

    

    /**
     * \brief The timeout when sending configs associated with this control
     */
    units::time::second_t ConfigTimeout{0.1_s};

    /**
     * \brief The period at which this control will update at.
     * This is designated in Hertz, with a minimum of 20 Hz
     * (every 50 ms) and a maximum of 1000 Hz (every 1 ms).
     *
     * If this field is set to 0 Hz, the control request will
     * be sent immediately as a one-shot frame. This may be useful
     * for advanced applications that require outputs to be
     * synchronized with data acquisition. In this case, we
     * recommend not exceeding 50 ms between control calls.
     */
    units::frequency::hertz_t UpdateFreqHz{20_Hz}; // Default to 20_Hz

    /**
     *\brief  Request coast neutral output of actuator.  The bridge is disabled and
     *        the rotor is allowed to coast.
     *
     */
    CoastOut() : ControlRequest{"CoastOut"}
    {}
    
    /**
     * \brief Sets the period at which this control will update at.
     * This is designated in Hertz, with a minimum of 20 Hz
     * (every 50 ms) and a maximum of 1000 Hz (every 1 ms).
     *
     * If this field is set to 0 Hz, the control request will
     * be sent immediately as a one-shot frame. This may be useful
     * for advanced applications that require outputs to be
     * synchronized with data acquisition. In this case, we
     * recommend not exceeding 50 ms between control calls.
     *
     * \param newUpdateFreqHz Parameter to modify
     * \returns Itself
     */
    CoastOut &WithUpdateFreqHz(units::frequency::hertz_t newUpdateFreqHz)
    {
        UpdateFreqHz = newUpdateFreqHz;
        return *this;
    }
    /**
     * Returns a string representation of the object.
     *
     * \returns a string representation of the object.
     */
    std::string ToString() const
    {
        std::stringstream ss;
        ss << "class: CoastOut" << std::endl;
        
        return ss.str();
    }

    /**
     * \brief Forces configs to be applied the next time this is used in a setControl.
     *        This is not necessary in the majority of cases, because Phoenix will make sure configs are
     *        properly set when they are not already set
     */
    void ForceApplyConfigs() { ApplyConfigsOnRequest = true; }
};

}
}
}

