/*
 * Copyright (C) Cross The Road Electronics.  All rights reserved.
 * License information can be found in CTRE_LICENSE.txt
 * For support and suggestions contact support@ctr-electronics.com or file
 * an issue tracker at https://github.com/CrossTheRoadElec/Phoenix-Releases
 */
#pragma once

#include "ctre/phoenix6/configs/Configs.hpp"
#include "ctre/phoenix6/controls/ControlRequest.hpp"
#include "ctre/phoenix6/networking/interfaces/Control_Interface.h"
#include <sstream>
#include "ctre/phoenix6/controls/TorqueCurrentFOC.hpp"
#include "ctre/phoenix6/controls/VelocityTorqueCurrentFOC.hpp"
#include <units/frequency.h>
#include <units/time.h>


namespace ctre {
namespace phoenix6 {
namespace controls {
namespace compound {

/**
 * This feature requires Phoenix Pro.
 *
 * Differential control with torque current average target and velocity
 * difference target.
 */
class Diff_TorqueCurrentFOC_VelocityTorqueCurrentFOC : public ControlRequest
{
    bool ApplyConfigsOnRequest{false};

    ctre::phoenix::StatusCode SendRequest(const char *network, uint32_t deviceHash, bool cancelOtherRequests, std::shared_ptr<ControlRequest> &req)
    {
        std::stringstream ss;
        auto& ref = requestReference.GetNameValues();
        ss << AverageRequest.ToString(); ref["AverageRequest"] = ss.str(); ss.str(std::string{});
        ss << DifferentialRequest.ToString(); ref["DifferentialRequest"] = ss.str(); ss.str(std::string{});

        if (req.get() != this)
        {
            auto const reqCast = dynamic_cast<Diff_TorqueCurrentFOC_VelocityTorqueCurrentFOC *>(req.get());
            if (reqCast != nullptr)
            {
                *reqCast = *this;
            }
            else
            {
                req = std::make_shared<Diff_TorqueCurrentFOC_VelocityTorqueCurrentFOC>(*this);
            }
        }

        
        std::string strs{ss.str()};
        c_ctre_phoenix6_requestConfigApply(network, deviceHash, ConfigTimeout.to<double>(), strs.c_str(), strs.length(), ApplyConfigsOnRequest);
        ApplyConfigsOnRequest = false;
        return c_ctre_phoenix6_RequestControlDiff_TorqueCurrentFOC_VelocityTorqueCurrentFOC(network, deviceHash, UpdateFreqHz.to<double>(), cancelOtherRequests, AverageRequest.Output.to<double>(), AverageRequest.MaxAbsDutyCycle.to<double>(), AverageRequest.Deadband.to<double>(), AverageRequest.OverrideCoastDurNeutral, DifferentialRequest.Velocity.to<double>(), DifferentialRequest.Acceleration.to<double>(), DifferentialRequest.FeedForward.to<double>(), DifferentialRequest.Slot, DifferentialRequest.OverrideCoastDurNeutral);
    }

public:
    /**
     * Average TorqueCurrentFOC request of the mechanism.
     */
    TorqueCurrentFOC AverageRequest;
    /**
     * Differential VelocityTorqueCurrentFOC request of the mechanism.
     */
    VelocityTorqueCurrentFOC DifferentialRequest;

    

    /**
     * \brief The timeout when sending configs associated with this control
     */
    units::time::second_t ConfigTimeout{0.1_s};

    /**
     * \brief The period at which this control will update at.
     * This is designated in Hertz, with a minimum of 20 Hz
     * (every 50 ms) and a maximum of 1000 Hz (every 1 ms).
     *
     * If this field is set to 0 Hz, the control request will
     * be sent immediately as a one-shot frame. This may be useful
     * for advanced applications that require outputs to be
     * synchronized with data acquisition. In this case, we
     * recommend not exceeding 50 ms between control calls.
     */
    units::frequency::hertz_t UpdateFreqHz{100_Hz}; // Default to 100_Hz

    /**
     *\brief  This feature requires Phoenix Pro.
     *       
     *        Differential control with torque
     *        current average target and velocity difference target.
     *
     *\param AverageRequest     Average TorqueCurrentFOC request of the mechanism.
     *\param DifferentialRequest     Differential VelocityTorqueCurrentFOC request
     *                               of the mechanism.
     */
    Diff_TorqueCurrentFOC_VelocityTorqueCurrentFOC(TorqueCurrentFOC AverageRequest, VelocityTorqueCurrentFOC DifferentialRequest) : ControlRequest{"Diff_TorqueCurrentFOC_VelocityTorqueCurrentFOC"},
        AverageRequest{std::move(AverageRequest)},
        DifferentialRequest{std::move(DifferentialRequest)}
    {}
    
    /**
     * \brief Modifies this Control Request's AverageRequest parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newAverageRequest Parameter to modify
     * \returns Itself
     */
    Diff_TorqueCurrentFOC_VelocityTorqueCurrentFOC& WithAverageRequest(TorqueCurrentFOC newAverageRequest)
    {
        AverageRequest = std::move(newAverageRequest);
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's DifferentialRequest parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newDifferentialRequest Parameter to modify
     * \returns Itself
     */
    Diff_TorqueCurrentFOC_VelocityTorqueCurrentFOC& WithDifferentialRequest(VelocityTorqueCurrentFOC newDifferentialRequest)
    {
        DifferentialRequest = std::move(newDifferentialRequest);
        return *this;
    }
    /**
     * \brief Sets the period at which this control will update at.
     * This is designated in Hertz, with a minimum of 20 Hz
     * (every 50 ms) and a maximum of 1000 Hz (every 1 ms).
     *
     * If this field is set to 0 Hz, the control request will
     * be sent immediately as a one-shot frame. This may be useful
     * for advanced applications that require outputs to be
     * synchronized with data acquisition. In this case, we
     * recommend not exceeding 50 ms between control calls.
     *
     * \param newUpdateFreqHz Parameter to modify
     * \returns Itself
     */
    Diff_TorqueCurrentFOC_VelocityTorqueCurrentFOC &WithUpdateFreqHz(units::frequency::hertz_t newUpdateFreqHz)
    {
        UpdateFreqHz = newUpdateFreqHz;
        return *this;
    }
    /**
     * Returns a string representation of the object.
     *
     * \returns a string representation of the object.
     */
    std::string ToString() const
    {
        std::stringstream ss;
        ss << "class: Diff_TorqueCurrentFOC_VelocityTorqueCurrentFOC" << std::endl;
        ss << "AverageRequest:" << std::endl;
        ss << "    Output: " << AverageRequest.Output.to<double>() << std::endl;
        ss << "    MaxAbsDutyCycle: " << AverageRequest.MaxAbsDutyCycle.to<double>() << std::endl;
        ss << "    Deadband: " << AverageRequest.Deadband.to<double>() << std::endl;
        ss << "    OverrideCoastDurNeutral: " << AverageRequest.OverrideCoastDurNeutral << std::endl;
        ss << "DifferentialRequest:" << std::endl;
        ss << "    Velocity: " << DifferentialRequest.Velocity.to<double>() << std::endl;
        ss << "    Acceleration: " << DifferentialRequest.Acceleration.to<double>() << std::endl;
        ss << "    FeedForward: " << DifferentialRequest.FeedForward.to<double>() << std::endl;
        ss << "    Slot: " << DifferentialRequest.Slot << std::endl;
        ss << "    OverrideCoastDurNeutral: " << DifferentialRequest.OverrideCoastDurNeutral << std::endl;
        return ss.str();
    }

    /**
     * \brief Forces configs to be applied the next time this is used in a setControl.
     *        This is not necessary in the majority of cases, because Phoenix will make sure configs are
     *        properly set when they are not already set
     */
    void ForceApplyConfigs() { ApplyConfigsOnRequest = true; }
};

}
}
}
}

